config_dict = dict(
    ResultsModelWiseResults_OLD=dict(lm=["aic", "bic", "centered_tss", "condition_number", "df_model", "df_resid", "ess",
                                        "f_pvalue", "fvalue", "k_constant", "llf", "mse_model", "mse_resid", "mse_total",
                                        "nobs", "rsquared", "rsquared_adj", "scale", "ssr", "uncentered_tss"],
                                     glm=["aic", "bic", "deviance", "df_model", "df_resid",
                                          "k_constant", "llf", "llnull", "null_deviance", "pearson_chi2", "scale",
                                          "nobs"],
                                     gee=["ctol", "df_model", "df_resid", "k_constant", "scale",
                                          "score_norm"]
                                     ),
    ResultsModelWiseResults=dict(lm=["aic", "df_model", "df_resid", "f_pvalue", "fvalue", "mse_model", "mse_resid", "mse_total",
                                     "rsquared", "rsquared_adj", "ssr"],
                                 glm=["aic", "df_model", "df_resid", "deviance", "llf", "llnull", "null_deviance",
                                      "pearson_chi2"],
                                 gee=["ctol", "df_model", "df_resid", "k_constant", "scale",
                                      "score_norm"],
                                 power=["ss"],
                                 pcorr=["r_", "r_prime", "t_", "p_", 'df']
                                 ),
    ResultsModelVariableWiseResults=dict(lm=["bse", "params", "pvalues", "tvalues"],
                                         glm=["bse", "params", "pvalues", "tvalues"],
                                         gee=["bse", "params", "pvalues", "tvalues"],
                                         lme=["bse", "params", "pvalues", "tvalues"]),
    VSVoxelOPS=dict(slice_count="20",
                    version=1.08),

)



import tempfile
import pyminc.volumes.factory as fc
import pyVS
try:
    data_path = pyVS.get_data('I34.mnc')
    h1 = fc.volumeFromFile(data_path)
    h1.closeVolume()
    h = fc.volumeLikeFile(data_path, '{0}/test.mnc'.format(tempfile.gettempdir()))
    h.closeVolume()
    print('MINC IO test successfull. ')
except Exception as e:
    print('Error in MINC IO. MINC file operations may fail. PLease make sure if you have installed latest minc-toolkit and sourced it. ')
    print(e)



from pyVS.pyVoxelStats.pyVoxelStats import pyVoxelStats

class FileReaderWriter:
    def __init__(self, file_type):
        self._file_type = file_type
        self._file_readers = {'minc': MincUtil, 'nifti': NiftiUtil}
        self._file_writers = {'minc': MincUtil, 'nifti': NiftiUtil}

    def read_file(self, image_file):
        return self._file_readers[self._file_type]()._load(image_file)

    def write_file(self, data, file_name, ref_file):
        return self._file_writers[self._file_type]()._save(data, file_name, ref_file)


class FileUtil(pyVoxelStats):
    def __init__(self):
        pyVoxelStats.__init__(self)

    def _load(self, file_name):
        pass

    def _save(self, data, file_name, ref_file):
        pass


class MincUtil(FileUtil):
    def __init__(self):
        FileUtil.__init__(self)

    def _load(self, file_name):
        import pyminc.volumes.factory as fc
        try:
            vol_h = fc.volumeFromFile(file_name)
            data = vol_h.getdata()
            vol_h.closeVolume()
        except Exception as e:
            raise Exception('Exception in file reading - {0} - {1}'.format(file_name, e))
        return data

    def _save(self, data, file_name, ref_file):
        import pyminc.volumes.factory as fc
        out_h = fc.volumeLikeFile(ref_file, file_name, volumeType='ushort')
        #out_h = fc.volumeLikeFile(ref_file, file_name)
        out_h.data = data
        out_h.writeFile()
        out_h.closeVolume()


class NiftiUtil(FileUtil):
    def __init__(self):
        FileUtil.__init__(self)

    def _load(self, file_name):
        import nibabel
        img = nibabel.load(file_name)
        return img.get_data()

    def _save(self, data, file_name, ref_file):
        import nibabel
        ref_img = nibabel.load(ref_file)
        affine = ref_img.affine
        nibabel.save(nibabel.Nifti1Image(data, affine), file_name)



import configparser
import os

import pyVS.Util.Params.pyVSParams as pyvsparams

class pyVoxelStats:
    _debug = False
    _no_parallel = False
    _save_model = False

    def __init__(self, file_type=None, model_string=None, csv_file=None, mask_file=None, voxel_variables=None, subset_string=None,
                 multi_variable_operations=None):
        self.config = self.check_config()
        self.string_model = model_string
        self.file_type = file_type
        self.data_file = csv_file
        self.mask_file = mask_file
        self.voxel_vars = voxel_variables
        self.filter_string = subset_string
        self.multi_var_operations = multi_variable_operations

        self.string_model_obj = None
        self.data_set = None

        self.string_model_obj = None
        self.data_set = None
        self.masker = None
        self.stats_model = None

        self.res = None
        self.models = None
        self.preds = None

        self.cluster_profile = 'default'
        self.clus_workers = None
        self.clus_no_start = False
        self.clust_sleep_time = 10
        self.clust_json = None

        self.cluster_shut_down = False

    def set_no_parallel(self, bool_val):
        pyVoxelStats._no_parallel = bool_val

    def enable_save_model(self):
        pyVoxelStats._save_model = True

    def set_debug(self, bool_val):
        pyVoxelStats._debug = bool_val

    def set_new_config(self, config_path):
        config_dict = pyvsparams.config_dict
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        cf = configparser.ConfigParser()
        for k in config_dict.keys():
            cf[k] = config_dict[k]
        with open(config_path, 'w') as configfile:
            cf.write(configfile)
        return cf

    def check_config(self, config_path='{0}/.pyVS/pyVSParams.ini'.format(os.path.expanduser('~'))):
        if os.path.exists(config_path):
            cf = configparser.ConfigParser()
            cf.read(config_path)
            config_dict = pyvsparams.config_dict
            if 'version' not in cf['VSVoxelOPS'] or (
                config_dict['VSVoxelOPS']['version'] > float(cf['VSVoxelOPS']['version'])):
                cf = self.set_new_config(config_path=config_path)
            return cf
        else:
            return self.set_new_config(config_path=config_path)

    def save(self, file_name, result_field, var_name=None):
        if var_name:
            print('Saving - {0} - {1} -> {2} ....'.format(result_field, var_name, file_name))
            self.masker.save_image_from_data(self.res[result_field][var_name], file_name)
        else:
            print('Saving - {0} -> {1} ....'.format(result_field, file_name))
            self.masker.save_image_from_data(self.res[result_field], file_name)
        print('Saving done.')

    def save_image_data(self, file_name, data):
        print('Saving - {0} ...'.format(file_name))
        self.masker.save_image_from_data(data, file_name)

    def set_up_cluster(self, clus_json=None, profile_name='default', workers=None, no_start=False, clust_sleep_time=10):
        self.cluster_profile = profile_name
        self.clus_workers = workers
        self.clus_no_start = no_start
        self.clust_sleep_time = clust_sleep_time
        self.clust_json = clus_json

    def set_shut_down_cluster(self):
        self.cluster_shut_down = True

    def evaluate(self):
        raise NotImplementedError()

    def cv_evaluate(self, cv_generator=None, repeats=1):
        raise NotImplementedError()



import pandas, re, numpy
import statsmodels.formula.api as smf
import statsmodels.tools.sm_exceptions as sme
from pyVS.pyVoxelStats.pyVoxelStats import pyVoxelStats
import numexpr, copy


class DataMatrix(pyVoxelStats):
    def __init__(self):
        pyVoxelStats.__init__(self)
        self._X = None
        self._Y = None


class Dataset(pyVoxelStats):
    def __init__(self, file_name, file_type='csv', delimiter=',', filter_string=None, string_model_obj=None):
        pyVoxelStats.__init__(self)
        self._file_name = file_name
        self._file_type = file_type
        self._delimiter = delimiter
        self._filter_string = filter_string
        self._data_table_full = self.__load_data()
        self.string_model_obj = string_model_obj
        self._data_table = self.__get_data_table()

    def __load_data_file(self):
        data_set = None
        if self._file_type == 'csv':
            data_set = pandas.read_csv(self._file_name, delimiter=self._delimiter)
        return data_set

    def __filter_data_file(self, data_set):
        return data_set.query(self._filter_string)

    def __load_data(self):
        data_set = self.__load_data_file()
        if self._filter_string:
            filtered_data = self.__filter_data_file(data_set)
        else:
            filtered_data = data_set
        return filtered_data

    def __get_data_table(self, used_vars=None):
        if not used_vars:
            used_vars = self.string_model_obj._used_vars
        return self._data_table_full[used_vars]

    def get_no_subjects(self):
        return len(self._data_table.index)


class StatsModel(pyVoxelStats):
    def __init__(self, type):
        pyVoxelStats.__init__(self)
        self._type = type
        self.mod = None
        self.obs_var_name = None
        self.save_models = False

        try:
            self.model_wise_results_names = [re.sub('\\[|\\]', '', s.strip().replace("'", '')) for s in
                                             self.config['ResultsModelWiseResults'][self._type].split(',')]
        except KeyError:
            self.model_wise_results_names = None
        try:
            self.var_wise_results_names = [re.sub('\\[|\\]', '', s.strip().replace("'", '')) for s in
                                           self.config['ResultsModelVariableWiseResults'][self._type].split(',')]
        except KeyError:
            self.var_wise_results_names = None

    def fit(self, data_frame):
        raise NotImplementedError()

    def predict(self, data_frame):
        raise NotImplementedError()

    def filter_result_statsmodels(self, result, model):
        result_f = {}
        for vard in self.model_wise_results_names:
            if result:
                result_f[vard] = getattr(result, vard)
            else:
                result_f[vard] = 0
        variable_names_in_model_op = model.exog_names
        for vard in self.var_wise_results_names:
            if result:
                result_f[vard] = {name: getattr(result, vard)[name] for name in variable_names_in_model_op}
            else:
                result_f[vard] = {name: 0 for name in variable_names_in_model_op}
        result_f['variable_names_in_model_op'] = variable_names_in_model_op
        result_f['model_wise_results_names'] = self.model_wise_results_names
        result_f['var_wise_results_names'] = self.var_wise_results_names
        if self.save_models:
            result_f['model'] = model
        return result_f


class LM(StatsModel):
    def __init__(self, string_model, weights=None):
        StatsModel.__init__(self, 'lm')
        self.string_model = string_model
        self.weights = weights

    def fit(self, data_frame):
        if self.weights:
            if isinstance(self.weights, str):
                mod = smf.wls(formula=self.string_model._string_model_str, data=data_frame, weights=data_frame[self.weights].values)
            elif isinstance(self.weights, numpy.ndarray) or isinstance(self.weights, float):
                mod = smf.wls(formula=self.string_model._string_model_str, data=data_frame, weights=self.weights)
            else:
                print('Weights either has to be column name, ndarray or a float. Reverting to OLS regression')
                mod = smf.ols(formula=self.string_model._string_model_str, data=data_frame)
        else:
            mod = smf.ols(formula=self.string_model._string_model_str, data=data_frame)
        self.mod = mod
        self.obs_var_name = mod.endog_names
        try:
            res = mod.fit()
        except (sme.PerfectSeparationError, sme.MissingDataError) as e:
            res = None
            # print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_statsmodels(res, mod)

    def predict(self, data_frame):
        if not self.mod:
            raise Exception('Model is not fitted. Please fit the model prior to using predict')
        else:
            X_vars = copy.deepcopy(self.string_model._used_vars)
            X_vars.remove(self.mod.endog_names)
            data_frame = data_frame[X_vars]
            preds = self.mod.fit().predict(data_frame)
        return preds

class RLM(StatsModel):
    def __init__(self, string_model):
        StatsModel.__init__(self, 'rlm')
        self.string_model = string_model

    def fit(self, data_frame):
        mod = smf.rlm(formula=self.string_model._string_model_str, data=data_frame)
        self.mod = mod
        try:
            res = mod.fit()
        except (sme.PerfectSeparationError, sme.MissingDataError) as e:
            res = None
            # print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_statsmodels(res, mod)


class GLM(StatsModel):
    def __init__(self, string_model, family_obj):
        StatsModel.__init__(self, 'glm')
        self.string_model = string_model
        self.family_obj = family_obj

    def fit(self, data_frame):
        mod = smf.glm(formula=self.string_model._string_model_str, data=data_frame, family=self.family_obj)
        self.mod = mod
        try:
            res = mod.fit()
        except (sme.PerfectSeparationError, sme.MissingDataError) as e:
            res = None
            # print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_statsmodels(res, mod)


class LME(StatsModel):
    def __init__(self, string_model, groups):
        StatsModel.__init__(self, 'lme')
        self.string_model = string_model
        self.groups = groups

    def fit(self, data_frame):
        mod = smf.gee(formula=self.string_model._string_model_str, data=data_frame, groups=self.groups)
        self.mod = mod
        try:
            res = mod.fit()
        except (sme.PerfectSeparationError, sme.MissingDataError) as e:
            res = None
            # print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_statsmodels(res, mod)



class GEE(StatsModel):
    def __init__(self, string_model, family_obj, groups, covariance_obj, time):
        StatsModel.__init__(self, 'gee')
        self.string_model = string_model
        self.family_obj = family_obj
        self.covariance_obj = covariance_obj
        self.time = time
        self.groups = groups

    def fit(self, data_frame):
        mod = smf.gee(formula=self.string_model._string_model_str, groups=self.groups, data=data_frame, family=self.family_obj,
                      cov_struct=self.covariance_obj)
        self.mod = mod
        try:
            res = mod.fit()
        except (sme.PerfectSeparationError, sme.MissingDataError) as e:
            res = None
            print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_statsmodels(res, mod)

class GAM(StatsModel):
    def __init__(self, string_model, family_str, method_str):
        StatsModel.__init__(self, 'gam')
        self.string_model = string_model
        self.family_obj = family_str
        self.method = method_str
        self.model_wise_results_names = []
        self.var_wise_results_names = []

    def fit(self, data_frame):
        import readline
        import rpy2.robjects as r
        from rpy2.robjects.packages import importr
        from rpy2.robjects import pandas2ri
        pandas2ri.activate()
        mgcv = importr('mgcv')
        family = r.r(self.family_obj) if self.family_obj else r.r('gaussian()')
        method_s = self.method if self.method else 'REML'
        opt = r.StrVector(['perf'])
        #ctrl = mgcv.control(maxiter=200)
        try:
            res = mgcv.gam(r.Formula(self.string_model._string_model_str), family=family, data=data_frame, method=method_s)
        except Exception as e:
            res = None
            #print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        self.mod = res
        return self.filter_result_gam(res, None)

    def filter_result_gam(self, result, model):
        import rpy2.robjects as r
        if not result:
            return None
        summary = r.r.summary(result)
        model_wise_names = ["deviance", "null.deviance", "df.null", "aic", "scale", "df.residual"]
        model_wise_names_sum = ["dev.expl", "r.sq"]
        vars_in_ptable = list(summary.rx2('p.table').dimnames[0])
        cols_in_ptable = list(summary.rx2('p.table').dimnames[1])
        vars_in_stable = list(summary.rx2('s.table').dimnames[0])
        cols_in_stable = list(summary.rx2('s.table').dimnames[1])
        result_f = {}
        variable_names_in_model_op = vars_in_ptable + vars_in_stable
        for vard in model_wise_names:
            self.model_wise_results_names.append(vard)
            if result:
                result_f[vard] = result.rx2(vard)[0]
            else:
                result_f[vard] = 0
        for vard in model_wise_names_sum:
            self.model_wise_results_names.append(vard)
            if result:
                result_f[vard] = summary.rx2(vard)[0]
            else:
                result_f[vard] = 0
        for vard_idx in range(len(cols_in_ptable)):
            vard_s = 'p.table.{0}'.format(cols_in_ptable[vard_idx])
            self.var_wise_results_names.append(vard_s)
            if result:
                result_f[vard_s] = {vars_in_ptable[name_idx]: summary.rx2('p.table').rx(name_idx+1, True)[vard_idx] for name_idx in range(len(vars_in_ptable))}
            else:
                result_f[vard_s] = {vars_in_ptable[name_idx]: 0 for name_idx in range(len(vars_in_ptable))}
        for vard_idx in range(len(cols_in_stable)):
            vard_s = 's.table.{0}'.format(cols_in_stable[vard_idx])
            self.var_wise_results_names.append(vard_s)
            if result:
                result_f[vard_s] = {vars_in_stable[name_idx]: summary.rx2('s.table').rx(name_idx + 1, True)[vard_idx] for name_idx in range(len(vars_in_stable))}
            else:
                result_f[vard_s] = {vars_in_stable[name_idx]: 0 for name_idx in range(len(vars_in_stable))}
        result_f['variable_names_in_model_op'] = variable_names_in_model_op
        result_f['model_wise_results_names'] = self.model_wise_results_names
        result_f['var_wise_results_names'] = self.var_wise_results_names
        return result_f

class Power(StatsModel):
    def __init__(self, string_model):
        StatsModel.__init__(self, 'power')
        self.string_model = string_model

    def get_expr_string(self, data_frame):
        s = self.string_model._string_model_str
        if 'std' in s:
            s = s.replace('std', str(data_frame[self.string_model._voxel_vars[0]].std()))
        if 'mean' in s:
            s = s.replace('mean', str(data_frame[self.string_model._voxel_vars[0]].mean()))
        if 'var' in s:
            s = s.replace('var', str(data_frame[self.string_model._voxel_vars[0]].var()))
        return s

    def fit(self, data_frame):
        string_expr = self.get_expr_string(data_frame)
        try:
            res = float(numexpr.evaluate(string_expr))
        except:
            res = None
            # print('Statistics exception; result for the voxel may be set to 0 : ' + str(e))
        return self.filter_result_power(res)

    def filter_result_power(self, result):
        result_f = {}
        for vard in self.model_wise_results_names:
            if result:
                result_f[vard] = numpy.clip(result,-5000,5000)
            else:
                result_f[vard] = 0
        result_f['model_wise_results_names'] = self.model_wise_results_names
        return result_f

class Pcorr(StatsModel):
    def __init__(self, string_model, var_y, var_x):
        StatsModel.__init__(self, 'pcorr')
        self.string_model = string_model
        self.var_y = var_y
        self.var_x = var_x

    def fit(self, data_frame):
        import patsy
        dmatrices  = patsy.dmatrices(self.string_model._string_model_str, data_frame)
        y = dmatrices[0]
        x = dmatrices[1]
        col_indxs = x.design_info.column_name_indexes
        var_x_idx = col_indxs[self.var_x]
        C = numpy.hstack([y, x])

        dof = C.shape[0] - 2 - (C.shape[1] - 2 - 1)  ### n-2-k k = number of variables except the two in question, -1 to remove intercept
        try:
            res = self.parr_corr_idx(C, 0+1, var_x_idx+1+1) ## We added the y column at the beginning. So, we have to increase the index
        except:
            res=None

        return self.filter_results_pcorr(res, dof)

    def parr_corr_idx(self, C, i, j):
        from scipy import stats, linalg
        i = i - 1
        j = j - 1

        C = numpy.asarray(C)
        p = C.shape[1]
        idx = numpy.ones(p, dtype=numpy.bool)
        idx[i] = False
        idx[j] = False
        beta_i = linalg.lstsq(C[:, idx], C[:, j])[0]
        beta_j = linalg.lstsq(C[:, idx], C[:, i])[0]

        res_j = C[:, j] - C[:, idx].dot(beta_i)
        res_i = C[:, i] - C[:, idx].dot(beta_j)

        corr = stats.pearsonr(res_i, res_j)[0]
        return corr

    def partial_corr(self, C):
        from scipy import stats, linalg
        C = numpy.asarray(C)
        p = C.shape[1]
        P_corr = numpy.zeros((p, p), dtype=numpy.float)
        for i in range(p):
            P_corr[i, i] = 1
            for j in range(i + 1, p):
                idx = numpy.ones(p, dtype=numpy.bool)
                idx[i] = False
                idx[j] = False
                beta_i = linalg.lstsq(C[:, idx], C[:, j])[0]
                beta_j = linalg.lstsq(C[:, idx], C[:, i])[0]

                res_j = C[:, j] - C[:, idx].dot(beta_i)
                res_i = C[:, i] - C[:, idx].dot(beta_j)

                corr = stats.pearsonr(res_i, res_j)[0]
                P_corr[i, j] = corr
                P_corr[j, i] = corr

        return P_corr

    def filter_results_pcorr(self, pcorr, dof):
        import scipy
        result_f = {}
        if pcorr:
            result_f['df'] = dof
            result_f['r_'] = pcorr
            result_f['r_prime'] = numpy.arctan(pcorr)
            result_f['t_'] = pcorr * numpy.sqrt((dof) / (1 - pcorr * pcorr))
            result_f['p_'] = (1 - scipy.special.stdtr(dof, numpy.abs(result_f['t_']))) * 2
        else:
            result_f['df'] = dof
            result_f['r_'] = 0
            result_f['r_prime'] = 0
            result_f['t_'] = 0
            result_f['p_'] = 0
        result_f['model_wise_results_names'] = self.model_wise_results_names
        return result_f

class StringModel(pyVoxelStats):
    def __init__(self, string_model_str, voxel_vars, multi_var_ops=None):
        pyVoxelStats.__init__(self)
        self._string_model_str = string_model_str
        self._voxel_vars = voxel_vars
        self._used_vars = self.__get_used_vars(self._string_model_str)
        self.multi_var_ops = multi_var_ops

    def __get_used_vars(self, string_model):
        all_strings = re.findall(r"[.Cs\(\w\)\w']+", string_model)
        all_strings = [re.sub(r'C\(([\w]+)\)', r'\1', st) for st in all_strings]
        all_strings = [re.sub(r's\(([\w]+)\)', r'\1', st) for st in all_strings]
        unique_vars = set(all_strings)
        return list(unique_vars)

    def add_to_cars(self, string):
        strs = self.__get_used_vars(string)
        for s in strs:
            if s not in self._used_vars:
                self._used_vars.append(s)

    def add_extra_used_vars(self, var_name):
        if isinstance(var_name, str):
            self._used_vars.append(var_name)



from setuptools import setup, find_packages
import json

with open('pyVS/pkg_info.json') as fp:
    _info = json.load(fp)

setup(
    # Application name:
    name="pyVoxelStats",

    # Version number (initial):
    version=_info['version'],

    # Application author details:
    author="Sulantha Mathotaarachchi",
    author_email="sulantha.ms@gmail.com",

    # Packages
    packages=find_packages(),

    package_data = {'pyVS': ['data/*', 'pkg_info.json']},

    # Include additional files into the package
    include_package_data=True,

    # Details
    url=_info['url'],

    #
    # license="LICENSE.txt",
    description="Python implementation of the VoxelStats toolbox for neuroimage analysis",

    # long_description=open("README.txt").read(),

    # Dependent packages (distributions)
    install_requires=[
        "nibabel", "pandas", "numpy", "scipy", "numexpr",
        "pyminc",
        "statsmodels",
        "rpy2",
        "ipyparallel",
    ],
)



from pyVS.Util.Masker import Masker
from pyVS.Util.VoxelOperation import VoxelOperation

from pyVS.Util.StatsUtil import Dataset, StringModel
from pyVS.Util.StatsUtil import Pcorr
from pyVS.pyVoxelStats.pyVoxelStats import pyVoxelStats

class pyVoxelStatsPCorr(pyVoxelStats):
    def __init__(self, file_type, model_string, csv_file, mask_file, voxel_variables, var_x, var_y, subset_string=None,
                 multi_variable_operations=None):
        pyVoxelStats.__init__(self, file_type, model_string, csv_file, mask_file, voxel_variables, subset_string,
                              multi_variable_operations)
        self.var_x = var_x
        self.var_y = var_y

    def evaluate(self):
        self.string_model_obj = StringModel(self.string_model, self.voxel_vars, self.multi_var_operations)
        self.data_set = Dataset(self.data_file, filter_string=self.filter_string,
                                string_model_obj=self.string_model_obj)
        self.masker = Masker(self.file_type, self.mask_file)
        self.stats_model = Pcorr(self.string_model_obj, self.var_y, self.var_x)
        self.stats_model.save_models = self._save_model
        voxel_op = VoxelOperation(self.string_model_obj, self.data_set, self.masker, self.stats_model)
        try:
            voxel_op.set_up_cluster(clus_json=self.clust_json, profile_name=self.cluster_profile, workers=self.clus_workers,
                                    no_start=self.clus_no_start, clust_sleep_time=self.clust_sleep_time)
            voxel_op.set_up()
            voxel_op.execute()
            self.res = voxel_op.results.get_results()
            self.models = voxel_op.results.get_models()
        finally:
            voxel_op.shut_down_cluster(self.cluster_shut_down)




import numpy

from pyVS.Util.FileUtil import FileReaderWriter
from pyVS.pyVoxelStats.pyVoxelStats import pyVoxelStats


class Masker(pyVoxelStats):
    def __init__(self, file_type, mask_file):
        pyVoxelStats.__init__(self)
        self._file_type = file_type
        self._mask_file = mask_file
        self._fileReaderWriter = FileReaderWriter(self._file_type)
        self._mask_array = self.__read_mask()
        self._mask_ndim = self._mask_array.ndim
        self._mask_shape = self._mask_array.shape

        self.highDimMask = None
        self.ref_file = None

    def __read_mask(self):
        mask_array = self._fileReaderWriter.read_file(self._mask_file)
        return mask_array > 0.9

    def __mask_data(self, data):
        dim_dif = data.ndim - self._mask_ndim
        need_new_ref = False
        if dim_dif > 0:
            time_rep_array = [1] * data.ndim
            for dim in range(dim_dif):
                time_rep_array[dim] = data.shape[dim]
            mask = numpy.tile(self._mask_array, tuple(time_rep_array))
            if ((self.highDimMask is not None) and (mask.ndim > self.highDimMask.ndim)) or (self.highDimMask is None):
                self.highDimMask = mask
                need_new_ref = True

        elif dim_dif == 0:
            mask = self._mask_array
            need_new_ref = False
        else:
            print('Mask dimensions and data dimensions mismatch. Mask dimensions are higher than data dimension')
            return None, None
        return data[mask], need_new_ref

    def __mask_image_data(self, image_data):
        masked_data, need_new_ref = self.__mask_data(image_data)
        return masked_data, need_new_ref

    def mask_image(self, image_file):
        image_data = self._fileReaderWriter.read_file(image_file)
        masked_data, need_new_ref = self.__mask_data(image_data)
        if need_new_ref:
            self.ref_file = image_file
        return masked_data

    def __rebuild_image_data(self, data):
        if self.highDimMask is None:
            new_image_data = numpy.zeros(self._mask_array.shape, dtype=data.dtype)
            new_image_data[self._mask_array] = data
        else:
            new_image_data = numpy.zeros(self.highDimMask.shape, dtype=data.dtype)
            new_image_data[self.highDimMask] = data
        return new_image_data

    def save_image_from_data(self, data, file_name):
        new_data = self.__rebuild_image_data(data)
        if not self.mask_file:
            self.mask_file = self._mask_file
        if not self.mask_file and not self.ref_file:
            raise Exception('Error in saving file. Ref file isnt properly set.')
        if self.ref_file:
            self._fileReaderWriter.write_file(new_data, file_name, self.ref_file)
        else:
            self._fileReaderWriter.write_file(new_data, file_name, self.mask_file)

    def get_data_from_image(self, image_file_name):
        return self._fileReaderWriter.read_file(image_file_name)

    def mask_image_data(self, image_data):
        return self.__mask_image_data(image_data)



import datetime
import os
import subprocess
import sys
import time
from multiprocessing import Pool

import ipyparallel as ipp
import numexpr
import numpy
import pandas

from pyVS.pyVoxelStats.pyVoxelStats import pyVoxelStats
from pyVS.pyVoxelStats.ShareObj import ShareObj

class VoxelOperation(pyVoxelStats):
    def __init__(self, string_model_obj, dataset_obj, masker_obj, stats_obj):
        pyVoxelStats.__init__(self)
        self.string_model_obj = string_model_obj
        self.dataset_obj = dataset_obj
        self.masker = masker_obj
        self.stats_obj = stats_obj
        self.voxel_var_data_map = {}
        self.operation_dataset = {}
        self.total_voxel_ops = None
        self.results = None
        self.rc = None
        self.par_view = None
        self.number_of_engines = 0

        self.temp_package = None
        self.predictions = None


    def set_up_cluster(self, clus_json=None, profile_name='default', workers=None, no_start=False, clust_sleep_time=10):
        if not pyVoxelStats._no_parallel:
            print('Setting up cluster .....')
            if clus_json:
                self.rc = ipp.Client(clus_json)
            else:
                if not no_start:
                    self.cluster_profile = profile_name
                    p0 = subprocess.Popen(['ipcluster stop --profile={0}'.format(profile_name)], shell=True)
                    time.sleep(5)
                    if workers:
                        str_args = ['ipcluster start --profile={1} -n {0}'.format(workers, profile_name)]
                    else:
                        str_args = ['ipcluster start --profile={0}'.format(profile_name)]
                    p = subprocess.Popen(str_args, shell=True)
                    if profile_name == 'default':
                        time.sleep(clust_sleep_time)
                    else:
                        time.sleep(clust_sleep_time)
                self.rc = ipp.Client(profile=str.encode(profile_name))
            self.par_view = self.rc.direct_view(targets='all')
            self.number_of_engines = len(self.par_view)
            print('Connected to {0} workers. '.format(self.number_of_engines))

    def shut_down_cluster(self, cluster_shut_down):
        if cluster_shut_down:
            if not pyVoxelStats._no_parallel:
                print("Shutting down cluster .....")
                p0 = subprocess.Popen(['ipcluster stop --profile={0}'.format(self.cluster_profile)], shell=True)
                time.sleep(6)
                print("Shutting down cluster complete.")
        else:
            print("Cluster shutdown is skipped")

    def set_up(self):
        print('Setting up voxel operations ... ')
        self.read_voxel_vars()
        self.set_up_data_for_op()

    def read_voxel_vars(self):
        print('File reading ...')
        total_files = 0
        pool = Pool()
        for var in self.string_model_obj._voxel_vars:
            list_of_files = self.dataset_obj._data_table[var]
            image_data_list = pool.map(self.masker.get_data_from_image, list_of_files)
            var_data_list = []
            for image_idx in range(len(image_data_list)):
                masked_data, need_new_ref = self.masker.mask_image_data(image_data_list[image_idx])
                if need_new_ref:
                    self.masker.ref_file = list_of_files[image_idx]
                var_data_list.append(masked_data)
            total_files += len(var_data_list)
            self.voxel_var_data_map[var] = numpy.vstack(var_data_list)
        pool.close()
        pool.join()
        print('Done. Total files read : {0}'.format(total_files))

    def __apply_voxel_ops(self, var_name, var_data):
        if not self.string_model_obj.multi_var_ops:
            return var_data
        data_dict = {var_name : var_data}
        for op in self.string_model_obj.multi_var_ops:
            try:
                ret_data = numexpr.evaluate(op, local_dict=data_dict)
                return ret_data
            except:
                pass
        return var_data

    def set_up_data_for_op(self):
        if not self.voxel_var_data_map:
            print('Voxel Data not read. Cannot continue.')
            return None
        else:
            for var in self.string_model_obj._used_vars:
                if var not in self.string_model_obj._voxel_vars:
                    self.operation_dataset[var] = dict(data=VarDataAccessPointer(self.__apply_voxel_ops(var, self.dataset_obj._data_table[var]), 1),
                                                       shape=self.dataset_obj._data_table[var].shape)
                else:
                    self.operation_dataset[var] = dict(
                        data=VarDataAccessPointer(self.__apply_voxel_ops(var, self.voxel_var_data_map[var]), self.voxel_var_data_map[var].shape[1]),
                        shape=self.voxel_var_data_map[var].shape)
        self.total_voxel_ops = max(
            self.operation_dataset[k]['shape'][1] if len(self.operation_dataset[k]['shape']) > 1 else 1 for k in
            self.operation_dataset)
        self.results = VoxelOpResultsWrapper(self.total_voxel_ops, self.stats_obj)
        self.results.save_model = self._save_model


    def __get_block_from_var_dict(self, block_variable_dict, start_loc):
        vars = list(block_variable_dict.keys())
        blockSize = block_variable_dict[vars[0]].shape[1]
        n_subs = block_variable_dict[vars[0]].shape[0]
        n_formats = ['f8'] * len(vars)
        n_dtype = dict(names=vars, formats=n_formats)
        n_data = numpy.zeros(n_subs, dtype=n_dtype)
        def f(k):
            for var in vars:
                n_data[var][:] = block_variable_dict[var][:, [k]].T
            return dict(data_block=pandas.DataFrame(n_data), location=start_loc + k, stats_obj=self.stats_obj)

        data_block = [f(x) for x in range(blockSize)]
        return data_block

    def __get_data_block(self, blockSize, block_number):
        finished = False
        if (blockSize * block_number) > self.total_voxel_ops:
            finished = True
            return ([], finished)
        elif (blockSize * (block_number + 1) > self.total_voxel_ops) & (
                blockSize * block_number < self.total_voxel_ops):
            block_var_dict = {}
            for var in self.operation_dataset:
                block_var_dict[var] = self.operation_dataset[var]['data'].get_data_block(int(blockSize * block_number),
                                                                                         int(self.total_voxel_ops - 1))
        else:
            block_var_dict = {}
            for var in self.operation_dataset:
                block_var_dict[var] = self.operation_dataset[var]['data'].get_data_block(int(blockSize * block_number),
                                                                                         int((blockSize * (
                                                                                         block_number + 1)) - 1))
        return (self.__get_block_from_var_dict(block_var_dict, int(blockSize * block_number)), finished)

    def __get_data_block_wIdx(self, blockSize, block_number, idx):
        finished = False
        if (blockSize * block_number) > self.total_voxel_ops:
            finished = True
            return ([], [], finished)
        elif (blockSize * (block_number + 1) > self.total_voxel_ops) & (
                        blockSize * block_number < self.total_voxel_ops):
            block_var_dict = {}
            for var in self.operation_dataset:
                block_var_dict[var] = self.operation_dataset[var]['data'].get_data_block_wIdx(int(blockSize * block_number),
                                                                                         int(self.total_voxel_ops - 1), idx)
        else:
            block_var_dict = {}
            for var in self.operation_dataset:
                block_var_dict[var] = self.operation_dataset[var]['data'].get_data_block_wIdx(int(blockSize * block_number),
                                                                                         int((blockSize * (
                                                                                             block_number + 1)) - 1), idx)
        return (self.__get_block_from_var_dict(block_var_dict, int(blockSize * block_number)), finished)

    def __get_data_block_cv(self, blockSize, block_number, train_idx, test_idx):
        train_blk, finished = self.__get_data_block_wIdx(blockSize, block_number, train_idx)
        test_blk, finished = self.__get_data_block_wIdx(blockSize, block_number, test_idx)
        return train_blk, test_blk, finished

    def execute(self):
        print('Execution started ... ')
        ex_st_time = datetime.datetime.now()
        slice_count = int(self.config['VSVoxelOPS']['slice_count'])
        print('Slices - {0}'.format(slice_count))
        blockSize = numpy.ceil(self.total_voxel_ops / slice_count)
        all_results = []
        for art_slice in range(slice_count):
            print('Slice - {0}/{1}'.format(art_slice + 1, slice_count))
            sl_st_time = datetime.datetime.now()
            data_block, finished = self.__get_data_block(blockSize, art_slice)
            d_end_time = datetime.datetime.now()
            if self._debug: print('Block creation time - {0}'.format(d_end_time-sl_st_time))
            if finished:
                print('Analysis complete')
            else:
                pr_st_time = datetime.datetime.now()
                if self._no_parallel:
                    par_results = map(run_par, data_block)
                else:
                    self.par_view.map(os.chdir, [os.getcwd()] * self.number_of_engines)
                    par_results = self.par_view.map_sync(run_par, data_block)
                pr_end_time = datetime.datetime.now()
                all_results.extend(par_results)
                ext_end_time = datetime.datetime.now()
                if self._debug: print('Parallel time - {0}'.format(pr_end_time - pr_st_time))
                if self._debug: print('Extend time - {0}'.format(ext_end_time - pr_end_time))
            sl_end_time = datetime.datetime.now()
            print(' - Time: {0} - Remaining time : {1}'.format((sl_end_time - sl_st_time), (sl_end_time - sl_st_time) * (slice_count - art_slice + 1)))
        self.results.temp_results = all_results
        full_end_time = datetime.datetime.now()
        print('Total execution time {0}'.format((full_end_time - ex_st_time)))

    def cv_execute(self, cv_generator=None, repeats=1):
        if not cv_generator:
            raise Exception('Please provide a scikit learn cross validation generator - Eg: KFold, LOO')
        if repeats > 1:
            print('Please make sure the cv generator uses shuffle when using repeats. ')

        print('Cross validation execution started ...')
        ex_st_time = datetime.datetime.now()
        slice_count = int(self.config['VSVoxelOPS']['slice_count'])
        print('Slices - {0}'.format(slice_count))
        blockSize = numpy.ceil(self.total_voxel_ops / slice_count)

        X = numpy.arange(self.dataset_obj.get_no_subjects())
        all_results = []
        all_preds = []
        all_obs = []
        for rep in range(repeats):
            print('In repeat {0}'.format(rep))
            cv_count = 0
            rep_results = []
            rep_preds = numpy.zeros((self.dataset_obj.get_no_subjects(), self.total_voxel_ops))
            rep_obs = numpy.zeros((self.dataset_obj.get_no_subjects(), self.total_voxel_ops))
            for train_idx, test_idx in cv_generator.split(X):
                cv_count += 1
                print('In CV - {0}'.format(cv_count))
                cv_k_results = []
                for art_slice in range(slice_count):
                    print('Slice - {0}/{1} ; CV - {2} ; Rep - {3}'.format(art_slice + 1, slice_count, cv_count, rep+1))
                    sl_st_time = datetime.datetime.now()
                    train_data_block, test_data_block, finished = self.__get_data_block_cv(blockSize, art_slice, train_idx, test_idx)
                    d_end_time = datetime.datetime.now()
                    if self._debug: print('Block creation time - {0}'.format(d_end_time - sl_st_time))
                    if finished:
                        print('Analysis complete for CV')
                    else:
                        self.par_view.map(os.chdir, [os.getcwd()] * self.number_of_engines)
                        #self.par_view.map(sys.path.append,
                        #                 ['/home/sulantha/PycharmProjects/pyVS'] * self.number_of_engines)
                        #self.par_view.map(sys.path.append,
                        #                  ['/home/sulantha/PycharmProjects/pyVS/Util'] * self.number_of_engines)
                        pr_st_time = datetime.datetime.now()
                        if self._no_parallel:
                            par_results_wPred = map(run_par_cv, train_data_block, test_data_block)
                        else:
                            par_results_wPred = self.par_view.map_sync(run_par_cv, train_data_block, test_data_block)
                        pr_end_time = datetime.datetime.now()
                        par_results, preds, obs = zip(*par_results_wPred)
                        par_results = list(par_results)
                        preds = list(preds)
                        obs = list(obs)
                        cv_k_results.extend(par_results)

                        block_locations = [t_blk_dict['location'] for t_blk_dict in test_data_block]
                        pred_res_counter = 0
                        for test_id in test_idx:
                            rep_preds[test_id, numpy.array(block_locations)] = numpy.array(preds).T[pred_res_counter, :]
                            rep_obs[test_id, numpy.array(block_locations)] = numpy.array(obs).T[pred_res_counter, :]
                            pred_res_counter += 1
                        ext_end_time = datetime.datetime.now()
                        if self._debug: print('Parallel time - {0}'.format(pr_end_time - pr_st_time))
                        if self._debug: print('Extend time - {0}'.format(ext_end_time - pr_end_time))
                    sl_end_time = datetime.datetime.now()
                    print(' - Time: {0} - Remaining time in this CV : {1}'.format((sl_end_time - sl_st_time),
                                                                       (sl_end_time - sl_st_time) * (
                                                                       slice_count - art_slice + 1)))

                rep_results.append(cv_k_results)
                print('CV - {0} Done. '.format(cv_count))
            all_results.append(rep_results)
            all_preds.append(rep_preds)
            all_obs.append(rep_obs)
            print('Rep - {0} Done.'.format(rep+1))

        self.results.temp_results = all_results
        self.predictions = CV_PredsHandler(repeats, all_preds, all_obs[0])
        full_end_time = datetime.datetime.now()
        print('Total execution time {0}'.format((full_end_time - ex_st_time)))

def ParGetBlock(i):
    data = {var: ShareObj.block_dict['block_variable_dict'][var][:, i] for var in ShareObj.block_dict['var_names']}
    return dict(data_block=pandas.DataFrame.from_dict(data), location=ShareObj.block_dict['start_loc'] + i,
                stats_obj=ShareObj.block_dict['stats_obj'])

def run_par(data_block):
    loc = data_block['location']
    res = data_block['stats_obj'].fit(data_block['data_block'])
    return ParRes(loc, res)

def run_par_cv(train_data_block, test_data_block):
    loc = train_data_block['location']
    res = train_data_block['stats_obj'].fit(train_data_block['data_block'])
    preds = train_data_block['stats_obj'].predict(test_data_block['data_block'])
    obs_var_name = train_data_block['stats_obj'].obs_var_name
    obs = test_data_block['data_block'][obs_var_name].as_matrix()
    return ParRes(loc, res), preds, obs

class ParRes:
    def __init__(self, loc, res):
        self.loc = loc
        self.res = res

class VarDataAccessPointer:
    def __init__(self, var_data, outer_shape):
        self.var_data = var_data
        self.outer_shape = outer_shape

    def get_data(self, pointer_loc):
        if self.outer_shape == 1:
            return self.var_data.to_numpy(copy=True)
        return self.var_data[:, pointer_loc % self.outer_shape]

    def get_data_block(self, loc_1, loc_2):
        data = [self.get_data(loc) for loc in range(loc_1, loc_2 + 1)]
        return numpy.vstack(data).T

    def get_data_block_wIdx(self, loc_1, loc_2, idx):
        data = [self.get_data(loc) for loc in range(loc_1, loc_2 + 1)]
        return numpy.vstack(data).T[idx]

class VoxelOpResultsWrapper:
    def __init__(self, total_voxel_operations, stats_model):
        self.total_ops = total_voxel_operations
        self.stats_model = stats_model
        self.temp_results = None
        self.__results = None

        self.__models = None

        self.save_model = False


    def modify_temp_result_parallel(self, result):
        self.modify_temp_result(result.res, result.loc)

    def modify_temp_result(self, value, loc):
        self.temp_results.insert(loc, value)

    def get_results(self):
        if not self.__results:
            self.__get_final_voxel_op_result()
        return self.__results

    def get_models(self):
        if not isinstance(self.__models, numpy.ndarray) and self.save_model:
            self.__get_final_voxel_op_result()
        return self.__models

    def get_model_vars_and_params(self):
        model_wise_results_names = None
        var_wise_results_names = None
        model_var_names = None
        result_good = False
        for o in self.temp_results: ## This section was added becuase of a problem arised from the GAM analysis. As we use rpy2, if there is an statical error, the result object is None and do not contain any info to
            # get the variable infomation. So we go through all the results to find the one which is not None to get the relavent information.
            if o.res:
                try:
                    model_var_names = o.res['variable_names_in_model_op']
                except KeyError:
                    model_var_names = None
                result_good = True
                try:
                    model_wise_results_names = o.res['model_wise_results_names']
                    var_wise_results_names = o.res['var_wise_results_names']
                except:
                    model_wise_results_names = self.stats_model.model_wise_results_names
                    var_wise_results_names = self.stats_model.var_wise_results_names
                if model_var_names:
                    var_wise_results_dict = {var_name: set() for var_name in var_wise_results_names}
                    for name in model_var_names:
                        for var_wise_name in var_wise_results_names:
                            if name in o.res[var_wise_name]:
                                var_wise_results_dict[var_wise_name].add(name)
                    var_wise_results_dict = {var_name: list(var_wise_results_dict[var_name]) for var_name in var_wise_results_dict}
                else:
                    var_wise_results_dict = None
                break
        if not result_good:
            return None, None, None, result_good

        if self.stats_model.model_wise_results_names:
            model_wise_results_names = self.stats_model.model_wise_results_names
            var_wise_results_names = self.stats_model.var_wise_results_names
            return model_wise_results_names, var_wise_results_names, model_var_names, var_wise_results_dict, result_good
        return model_wise_results_names, var_wise_results_names, model_var_names, var_wise_results_dict, result_good

    def __get_final_voxel_op_result(self):
        print('Building final results. This may take a while depending on the dimensions of the images. ')
        res_bl_st = datetime.datetime.now()
        model_wise_results_names, var_wise_results_names, model_var_names, var_wise_results_dict, results_good = self.get_model_vars_and_params()
        if results_good:
            builer = ResultBuilder(self.temp_results, self.total_ops, model_wise_results_names, var_wise_results_names, model_var_names, var_wise_results_dict, results_good)
            builer.save_model = self.save_model
            self.__results, self.__models = builer.make_result()
            self.temp_results = None
            print('Final results building finished. ')
        else:
            print('Error in results. None of the results were successful. Check statistical errors. Use no parallel option in VoxelOperation to see the errors at voxel level.')
            self.temp_results = None
            self.results = None
        res_bl_end = datetime.datetime.now()
        print('Time taken to build final results: {0}'.format(res_bl_end-res_bl_st))


class ResultBuilder:
    def __init__(self, temp_results, total_ops, model_wise_results_names, var_wise_results_names, model_var_names, var_wise_results_dict, results_good):
        self.temp_results = temp_results
        self.total_ops = total_ops
        self.model_wise_results_names = list(set(model_wise_results_names)) if model_wise_results_names else None
        self.var_wise_results_names= list(set(var_wise_results_names)) if var_wise_results_names else None
        self.model_var_names = list(set(model_var_names)) if model_var_names else None
        self.var_wise_results_dict = var_wise_results_dict
        self.results_good = results_good
        self.save_model = False

    def value_to_record(self, value):
        tup = ()
        if self.model_wise_results_names:
            tup = tup + tuple(value.res[var] for var in self.model_wise_results_names)
        if self.var_wise_results_names:
            tup = tup + tuple(tuple(value.res[var][name] for name in self.var_wise_results_dict[var]) for var in self.var_wise_results_dict)
        return tup

    def make_result(self):
        self.temp_results.sort(key=lambda x: x.loc, reverse=False)
        dtype = []
        if self.model_wise_results_names:
            for var in self.model_wise_results_names:
                dtype.append(("{0}".format(var), "f8"))
        if self.var_wise_results_names:
            for var in self.var_wise_results_dict:
                dtype.append(("{0}".format(var), [("{0}".format(name), "f8") for name in self.var_wise_results_dict[var]]))
        print('Outputs: Model wise: {0}'.format(self.model_wise_results_names))
        print('Outputs:  Variable wise: {0}'.format(self.var_wise_results_names))
        print('Variables names in model: Model wise: {0}'.format(self.model_var_names))
        arr = numpy.fromiter(map(self.value_to_record, self.temp_results), dtype=dtype, count=self.total_ops)
        models = None
        if self.save_model:
            models = numpy.array([t_res.res['model'] for t_res in self.temp_results])
        return arr, models

    def make_result_p(self, obj, res):
        i = obj.loc
        result = obj.res
        if  self.model_wise_results_names:
            for var in self.model_wise_results_names:
                res[var][i] = result[var]
        if  self.var_wise_results_names:
            for var in self.var_wise_results_names:
                for name in self.model_var_names:
                    try:
                        res[var][name][i] = result[var][name]
                    except KeyError as e:
                        ##TODO Add a debug check
                        pass

        return None

class CV_PredsHandler:
    def __init__(self, n_repeats, predictions, obs_data):
        self.n_repeats = n_repeats
        self.preds = numpy.array(predictions)
        self.obs_data = numpy.array(obs_data)

    def get_predictions(self):
        return self.preds

    def get_mean_prediction(self):
        return numpy.mean(self.preds, 0)

    def get_error_pred(self):
        return self.preds - self.obs_data

    def get_rmse(self):
        return numpy.sqrt(numpy.mean(numpy.square(self.get_mean_prediction() - self.obs_data), axis=0))



import os
import json

_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    return os.path.join(_ROOT, 'data', path)

with open(os.path.dirname(__file__) + '/pkg_info.json') as fp:
    _info = json.load(fp)

__version__ = _info['version']
